package org.sc3d.apt.sss.v3;

/** Represents an indentation analysis of a Sentence. */
public class Indentation {
  /** Constructs an Indentation analysis of 'lex.sentence', and annotates 'lex.sentence' with error messages if the indentation rules are not obeyed. */
  public Indentation(Lex lex) {
    this.sentence = lex.sentence;
    // Find the non-blank lines and count their indentation. */
    final IntBuffer starts = new IntBuffer(), indents = new IntBuffer();
    int j = 0; // Index of first token that ends after 'i'.
    for (int i=0; i<this.sentence.length; ) {
      // Count the indentation.
      int start = i, pos = 0;
      while (i<this.sentence.length) {
        final char c = this.sentence.get(i++);
        if (c=='\n') { start = i; pos = 0; }
        else if (c=='\r') pos = 0;
        else if (c==' ') pos++;
        else if (c=='\t') pos = -(-8&~pos);
        else break;
      }
      // Check it's not inside a Token.
      while (j<lex.numTokens) {
        final Token t = lex.getToken(j);
        if (t.start>=start && t.type!=Token.TYPE_COMMENT) {
          starts.append(start);
          indents.append(pos);
        }
        if (t.start+t.length>start) break;
        j++;
      }
      // Find the next new-line.
      while (i<this.sentence.length && this.sentence.get(i)!='\n') i++;
    }
    this.starts = starts.toArray();
    this.indents = indents.toArray();
    this.numLines = this.starts.length;
    // Work out the depths.
    this.depths = new int[this.numLines];
    int currentDepth = 0, minDepth = 0, maxDepth = 0;
    j = 0;
    for (int i=0; i<this.numLines; i++) {
      final int start = this.getLineStart(i);
      final int end =
        i+1<this.numLines ? this.getLineStart(i+1) : this.sentence.length;
      this.depths[i] = currentDepth;
      while (j<lex.numTokens) {
        final Token t = lex.getToken(j);
        if (t.start>=end) break;
        if (t.length==1) {
          final char c = this.sentence.get(t.start);
          if (c=='(' || c=='[' || c=='{') currentDepth++;
          else if (c==')' || c==']' || c=='}') currentDepth--;
          if (currentDepth<this.depths[i]) this.depths[i] = currentDepth;
        }
        j++;
      }
      if (this.depths[i]<minDepth) minDepth = this.depths[i];
      else if (this.depths[i]>maxDepth) maxDepth = this.depths[i];
    }
    // Do local indentation checks. These checks are incomplete, but produce
    // decent error messages even if the bracket matching is rubbish.
    boolean okay = true;
    for (int i=1; i<this.numLines; i++) {
      final int dp = this.getLineIndentation(i) - this.getLineIndentation(i-1);
      final int dd = this.getLineDepth(i) - this.getLineDepth(i-1);
      if (dd>0) {
        okay &= dp>0;
        if (dp<=0) this.addIndentationError(
          "This line is enclosed in more brackets or braces than the previous "+
          "line, and must therefore be indented more.",
          i
        );
      } else if (dd<0) {
        okay &= dp<0;
        if (dp>=0) this.addIndentationError(
          "This line is enclosed in more brackets or braces than the next "+
          "line, and must therefore be indented more.",
          i-1
        );
      } else { // 'dd==0'.
        okay &= dp==0;
        if (dp<0) this.addIndentationError(
          "This line is enclosed in the same number of brackets or braces as "+
          "the previous line, and must therefore be indented by the same "+
          "amount. Perhaps you have forgotten to close a bracket or brace?",
          i
        );
        if (dp>0) this.addIndentationError(
          "This line is enclosed in the same number of brackets or braces as "+
          "the next line, and must therefore be indented by the same "+
          "amount. Perhaps you have forgotten to open a bracket or brace?",
          i-1
        );
      }
    }
    // If everything looks okay so far, do the full indentation check.
    if (okay) {
      // 'lines[i]' is the most recent line with depth at most 'minDepth+i'.
      final int[] lines = new int[maxDepth-minDepth+1];
      for (int i=0; i<this.numLines; i++) {
        final int pos = this.getLineIndentation(i);
        final int depth = this.getLineDepth(i);
        if (depth==minDepth && pos!=0) this.addIndentationError(
          "This line is not enclosed in any brackets or braces "+
          "and must therefore not be indented.",
          i
        );
        // Compare indentation to a deeper line.
        for (int d=depth+1; d<maxDepth; d++) {
          final int k = lines[d-minDepth];
          if (depth<this.getLineDepth(k)) {
            if (pos>=this.getLineIndentation(k)) this.addIndentationError(
              "This line is enclosed in more brackets or braces than "+
              "a later line and must therefore be indented more than "+
              pos+" spaces. "+
              "Perhaps there is an open bracket or brace missing?",
              k
            );
            break;
          }
        }
        // Compare to a shallower line or a line at the same depth.
        final int k = lines[depth-minDepth];
        if (depth>this.getLineDepth(k)) {
          if (pos<=this.getLineIndentation(k)) this.addIndentationError(
            "This line is enclosed in more brackets or braces than "+
            "an earlier line and must therefore be indented more than "+
            this.getLineIndentation(k)+" spaces. "+
            "Perhaps there is a close bracket or brace missing?",
            i
          );
        } else { // 'depth==this.getLineDepth(k)'.
          if (pos!=this.getLineIndentation(k)) this.addIndentationError(
            "This line is enclosed in the same number of brackets or braces "+
            "as an earlier line and must therefore be indented by "+
            this.getLineIndentation(k)+" spaces. "+
            "Perhaps there is a bracket matching error?",
            i
          );
        }
        // Update 'lines'.
        for (int d=depth; d<maxDepth; d++) lines[d-minDepth] = i;
      }
    }
  }
  
  /* New API. */
  
  /** The Sentence of which this is the indentation analysis. */
  public final Sentence sentence;
  
  /** The number of lines in 'sentence' to which the indentation rules apply. Lines consisting entirely of white-space don't count. Lines which start inside a comment or string literal don't count. */
  public final int numLines;
  
  /** Returns the index into 'sentence' of the first character of a line.
   * @param line the line number. Lines are numbered from '0' to 'numLines-1'. This numbering differs from that used in error messages, which is chosen to match the behaviour of text editors.
   * @return the index of the first character of the specified line.
   */
  public int getLineStart(int line) { return this.starts[line]; }
  
  /** Returns the indentation of a line. The indentation is represented as the horizontal position of the text cursor after printing the initial white-space characters of the line using the normal unix rules. To be specific, ' ' moves the cursor one place, '\t' moves it at least one place to the next multiple of eight, and '\r' returns the cursor to the beginning of the line.
   * @param line the line number. Lines are numbered from '0' to 'numLines-1'.
   * @return the number of spaces equivalent to the indentation.
   */
  public int getLineIndentation(int line) { return this.indents[line]; }
  
  /** Returns the nesting depth of a line. According to the SSS specification, the nesting depth is defined to be the number of brackets or braces that completely enclose the line, including the new-line characters before and after it, if any.
   * <p>This method actually uses a different definition which coincides with the proper one when the brackets match correctly. The depth of a line is the smallest depth of any of the characters on the line. The depth of a character is the number of open brackets, open square brackets and open braces before it minus the number of close brackets, close square brackets and close braces before it.
   * @param line the line number, from '0' to 'numLines-1'.
   * @return the nesting depth.
   */
  public int getLineDepth(int line) { return this.depths[line]; }
  
  /** Attaches an error message to 'sentence' about the indentation at the beginning of line 'line'. Use this method only for errors about indentation. */
  public void addIndentationError(String message, int line) {
    this.sentence.addError(
      message,
      this.getLineStart(line),
      this.getLineIndentation(line)
    );
  }
  
  /* Private. */
  
  /** An array of length 'numLines'. Element 'i' is the index of the first character of line 'i'. */
  private final int[] starts;
  
  /** An array of length 'numLines'. Element 'i' is the indentation of line 'i'. */
  private final int[] indents;
  
  /** An array of length 'numLines'. Element 'i' is the nesting depth of line 'i'. */
  private final int[] depths;
  
  /* Test code. */
  
  /** Takes one argument: a filename. */
  public static void main(String[] args) throws java.io.IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.sss.v3.Indentation <filename>"
    );
    final Sentence sentence = Sentence.readFile(args[0]);
    final Lex lex = new Lex(sentence);
    final Indentation me = new Indentation(lex);
    if (sentence.countErrors()!=0) {
      sentence.printErrorReport(System.out, 1000000);
    } else {
      System.out.println("No errors.");
    }
    System.out.println(me.numLines+" lines");
    for (int i=0; i<me.numLines; i++) {
      System.out.println(
        "Line "+i+
        " starts at "+me.getLineStart(i)+
        ", has indentation "+me.getLineIndentation(i)+
        " and has depth "+me.getLineDepth(i)
      );
    }
  }
}
