package org.sc3d.apt.sss.v3;

/** Represents the parse-tree of a Sentence. This is the return type of 'Parser.parseTokens()'. Its structure closely mimics that of 'Grammar'. 'null' represents a Sentence that contains errors. */
public abstract class Tree {
  /** Don't want any instances except of the inner classes. */
  private Tree(boolean isTerminal) { this.isTerminal = isTerminal; }
  
  /* New API. */
  
  /** 'true' if this Tree is a single Token corresponding to a Terminal of the grammar, otherwise 'false'. */
  public final boolean isTerminal;
  
  /** Returns a String representation of this Tree. */
  public String toString() {
    final StringBuffer ans = new StringBuffer();
    this.dump(ans, "");
    return ans.toString();
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** The subclass of Tree with 'isTerminal==true', which represents a single Token. In other words, this class represents a string that matches a Grammar.Terminal. */
  public static class Terminal extends Tree {
    /** Constructs a Terminal. */
    public Terminal(Token t) { super(true); this.t = t; }
    
    /** The Token that the parser found. */
    public Token t;
    
    /** If this Terminal represents a bracket, parses the contents and returns the parse-tree. If there are any errors, this method attaches appropriate error messages to the sentence and returns 'null'. Do not call this method if this Terminal does not represent a bracket. The default implementation throws a RuntimeException. */
    public Tree parse() {
      throw new RuntimeException("This Terminal does not represent a bracket");
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** The subclass of Tree with 'isTerminal==false', which represents zero or more parse-trees all matching the same non-terminal of the grammar. In other words, this class represents a string that matches a Grammar.NonTerminal. */
  public static class NonTerminal extends Tree {
    /** Constructs a NonTerminal. */
    public NonTerminal(Production[] repeats) {
      super(false);
      this.repeats = repeats; this.length = repeats.length;
    }
    
    /** The number of repeats that the parser found. If the Grammar.NonTerminal was marked as optional, this field may be zero, otherwise it must be non-zero. If the Grammar.NonTerminal was marked as repeatable, this field may be more than one, otherwise it must be at most one. */
    public final int length;
    
    /** Returns the parse-tree of the repeat with index 'index'. */
    public Production get(int index) { return this.repeats[index]; }
    
    // Following spec. is probably a mistake: should not ever return 'null'.
    
    /** Returns 'get(0)', or 'null' if 'length' is zero.
     * @throws IllegalArgumentException if 'length' is greater than one.
     */
    public Production get() {
      if (this.length==0) return null;
      if (this.length==1) return this.get(0);
      throw new IllegalArgumentException("Which one?");
    }
    
    private Production[] repeats;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** Represents something constructed using a Grammar.Production from a sequence of smaller things. */
  public static class Production {
    /** Constructs a Production. */
    public Production(String name, Tree[] parts) {
      this.name = name;
      this.parts = parts; this.numParts = parts.length;
    }
    
    /** The name of the Grammar.Production that was used. */
    public final String name;
    
    /** The number of parts. This is the same as the corresponding field of the Grammar.Production that was used. */
    public final int numParts;
    
    /** Returns the part with index 'index'. */
    public Tree get(int index) { return this.parts[index]; }
    
    /** Returns '(Terminal)get(index)'. */
    public Terminal getT(int index) { return (Terminal)this.get(index); }
    
    /** Returns '(NonTerminal)get(index)'. */
    public NonTerminal getNT(int index) { return (NonTerminal)this.get(index); }
    
    /** Returns '(Production)getNT(index).get()'. */
    public Production getP(int index) {
      return (Production)this.getNT(index).get();
    }
    
    /** Returns a String representation of this Production. */
    public String toString() {
      final StringBuffer ans = new StringBuffer("Tree.Production[");
      ans.append("name=").append(this.name);
      for (int i=0; i<this.numParts; i++)
        ans.append(", get(").append(i).append(")=").append(this.get(i));
      return ans.append("]").toString();
    }
    
    private final Tree[] parts;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /* Private. */

  /** Used to implement 'toString()'. Appends a String representation of this Tree to 'sb', prefixing every line with 'prefix'. */
  private void dump(StringBuffer sb, String prefix) {
    if (this.isTerminal) {
      sb.append(prefix).append("\\_").append(((Terminal)this).t).append('\n');
    } else {
      final NonTerminal nt = (NonTerminal)this;
      sb.append(prefix).append("\\ ").append(nt.length);
      sb.append(nt.length==1 ? " item\n" : " items\n");
      for (int i=0; i<nt.length; i++) {
        final String newPrefix = prefix + (i==nt.length-1 ? "  " : " |");
        final Production p = nt.get(i);
        sb.append(newPrefix).append("\\ ").append(p.name).append('\n');
        for (int j=0; j<p.numParts; j++) {
          p.get(j).dump(sb, newPrefix + (j==p.numParts-1 ? "  " : " |"));
        }
      }
    }
  }
}
